/*
 *   bpf_next.c: The main program of the loadable kernel server
 *               for the Berkeley Packet Filter.
 *
 *   Copyright (C) 1994 Satoshi Adachi.
 *
 *   This program is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by the
 *   Free Software Foundation; either version 2, or (at your option) any
 *   later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   --------------------------------------------------------------------------
 *
 *   REMARK:
 *   The promiscuous mode and the multicast address for the Ethernet driver
 *   have not yet been *officially* supported at least until NEXTSTEP3.2J.
 *   Actually, they exist in NEXTSTEP 3.0J and 3.2J, as they are used in
 *   this source code. We hope that the next release of NEXTSTEP will
 *   support it!
 *
 */

static char rcsid[] = "$Id: bpf_next.c,v 0.2 1994/08/10 17:19:57 adachi Exp adachi $";

/* ------------------------------------------------------------------------- */
/*   Include files.                                                          */
/* ------------------------------------------------------------------------- */

#include <sys/param.h>
#include <sys/errno.h>
#include <sys/printf.h>
#include <sys/conf.h>

#include <syslog.h>
#include <kernserv/kern_server_types.h>
#include <kernserv/prototypes.h>

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <net/netif.h>

#define LOG_DEBUG_SPECIAL LOG_DEBUG

#include "net/bpf.h"

/* ------------------------------------------------------------------------- */
/*   Declare private variables and functions.                                */
/* ------------------------------------------------------------------------- */

/* ----- Special file side ----- */

kern_server_t instance;

extern int nulldev();
extern int nodev();
#define nullftr 0

#define BPF_RAW_MAJOR 32

extern int bpfopen();
extern int bpfclose();
extern int bpfread();
extern int bpfwrite();
extern int bpfioctl();
extern int bpfselect();

static struct cdevsw bpf_cdevsw = {
  bpfopen,
  bpfclose,
  bpfread,
  bpfwrite,
  bpfioctl,
  nodev,
  nulldev,
  bpfselect,
  nodev,
  nodev,
  nodev
  };

static struct cdevsw bpf_saved_cdevsw;

/* ----- Network side ----- */

struct bpf_next_private {
  netif_t ifp;
  netif_t rifp;
  caddr_t sc_bpf;
  struct bpf_next_private *next;
};
typedef struct bpf_next_private bpf_next_private_t;

static bpf_next_private_t* bpf_next_private_list = NULL;

#define BPF_NEXT_RIF(ifp) (((bpf_next_private_t *)if_private(ifp))->rifp)

static void bpf_next_config();
static void bpf_next_attach(void *private, netif_t rifp);
static int bpf_next_input(netif_t ifp, netif_t rifp, netbuf_t nb, void *extra);
static void bpf_next_detach();

void bpf_next_log(int log_level, char *msg,
		  int arg1, int arg2, int arg3, int arg4, int arg5);

#define LOG_DEBUG_SPECIAL2 (LOG_DEBUG_SPECIAL + 1)
#define LOG_DEBUG_SPECIAL3 (LOG_DEBUG_SPECIAL + 2)

#if !defined(USE_PRINTF) && !defined(USE_KERN_SERV_LOG) && !defined(USE_LOG)
#define USE_KERN_SERV_LOG
#endif

/*****************************************************************************/
/*   bpf_next_init:                                                          */
/*****************************************************************************/

void
bpf_next_init()
{
  bpf_next_log(LOG_NOTICE, "bpf loaded\n", 0, 0, 0, 0, 0);

  /* Configure the network sniffers. */
  bpf_next_config();

  /* Save whatever entries were in the table for our major number. */
  bpf_saved_cdevsw = cdevsw[BPF_RAW_MAJOR];

  /* Put our entry in the switch table. */
  cdevsw[BPF_RAW_MAJOR] = bpf_cdevsw;
}

/*****************************************************************************/
/*   bpf_next_signoff:                                                       */
/*****************************************************************************/

void
bpf_next_signoff()
{
  /* Restore the saved entry in the switch table. */
  cdevsw[BPF_RAW_MAJOR] = bpf_saved_cdevsw;

  /* Do the termination for the network side. */
  bpf_next_detach();

  bpf_next_log(LOG_NOTICE, "bpf unloaded\n", 0, 0, 0, 0, 0);
}

/*****************************************************************************/
/*   bpf_next_config:                                                        */
/*****************************************************************************/

static void
bpf_next_config()
{
  if_registervirtual(bpf_next_attach, NULL);
}

/*****************************************************************************/
/*   bpf_next_attach:                                                        */
/*****************************************************************************/

static void
bpf_next_attach(private, rifp)
     void *private;
     netif_t rifp;
{
  netif_t ifp;
  bpf_next_private_t* ifprivate;
  const char *name;
  unsigned int unit;
  const char *type;
  unsigned int mtu;
  unsigned int flags;
  netif_class_t class;

  bpf_next_log(LOG_DEBUG_SPECIAL, "bpf_next_attach entered\n", 0, 0, 0, 0, 0);

  /* We have interest in only 10MB Ethernet. */
  if (strcmp(if_type(rifp), IFTYPE_ETHERNET) != 0) {
    return;
  }

  /* Wire down a private data in the kernel. */
  ifprivate = (bpf_next_private_t *)kalloc(sizeof(bpf_next_private_t));

  /* Attach the bpf network driver. */
  bpfattach(&ifprivate->sc_bpf, rifp, DLT_EN10MB, sizeof(struct ether_header));

  /* Set parameters for the sniffer. */
  name  = if_name(rifp);
  unit  = if_unit(rifp);
  type  = if_type(rifp);
  mtu   = if_mtu(rifp);
  flags = if_flags(rifp);
  class = if_class(rifp);
  
  /* Attach the module. */
  ifp = if_attach(NULL,                   /* init func */
		  bpf_next_input,         /* input func */
		  NULL,                   /* output func */
		  NULL,                   /* getbuf func */
		  NULL,                   /* control func */
		  name, unit, type,
		  mtu, flags,
		  NETIFCLASS_SNIFFER, ifprivate);

  /* Fill other fields in the private data. */
  ifprivate->ifp  = ifp;
  ifprivate->rifp = rifp;

  /* Link the private data to the global list. */
  ifprivate->next =  bpf_next_private_list;
  bpf_next_private_list = ifprivate;

  bpf_next_log(LOG_DEBUG_SPECIAL,
	       "bpf enabled for interface %s%d, type = \"%s\", mtu = %d, flags = 0x%x\n",
	       (int)name, (int)unit, (int)type, (int)mtu, (int)flags);
  bpf_next_log(LOG_DEBUG_SPECIAL, "class = %s\n",
	       (int)((class == NETIFCLASS_REAL) ? "real" :
		     ((class == NETIFCLASS_VIRTUAL) ? "virtual" :
		      ((class == NETIFCLASS_SNIFFER) ? "sniffer" : "unknown"))),
	       0, 0, 0, 0);

  bpf_next_log(LOG_DEBUG_SPECIAL, "bpf_next_attach leaving\n", 0, 0, 0, 0, 0);

  return;
}

/****************************************************************************/
/*   bpf_next_detach:                                                        */
/****************************************************************************/

static void
bpf_next_detach()
{
  bpf_next_private_t *p, *q;

  bpf_next_log(LOG_DEBUG_SPECIAL, "bpf_next_detach entered\n", 0, 0, 0, 0, 0);

  p = bpf_next_private_list;
   while (p) {
    q = p->next;
    if (p->ifp)
      if_detach(p->ifp);
    kfree(p, sizeof(bpf_next_private_t));
    p = q;
  }

  bpf_next_log(LOG_DEBUG_SPECIAL, "bpf_next_detach leaving\n", 0, 0, 0, 0, 0);

  return;
}

/*****************************************************************************/
/*   bpf_next_input:                                                         */
/*****************************************************************************/

static int
bpf_next_input(ifp, rifp, nb, extra)
     netif_t ifp;
     netif_t rifp;
     netbuf_t nb;
     void *extra;
{
  bpf_next_private_t *private = (bpf_next_private_t *)if_private(ifp);

  bpf_next_log(LOG_DEBUG_SPECIAL3, "bpf_next_input entered\n", 0, 0, 0, 0, 0);

  /* Do we want packets from this driver ? */
  if (private->rifp != rifp)
    return EAFNOSUPPORT;

  /* If the special file is open, then transfer the packet. */
  if (private->sc_bpf) {
    char * buf = nb_map(nb);
    unsigned int size = nb_size(nb);
    bpf_next_log(LOG_DEBUG_SPECIAL3, "bpf_next_input: calling bpf_tap ...\n",
		 0, 0, 0, 0, 0);
    bpf_tap(private->sc_bpf, buf, size);
  }

  bpf_next_log(LOG_DEBUG_SPECIAL3, "bpf_next_input leaving\n", 0, 0, 0, 0, 0);

  /* Return with the mark that packet is only sniffed. */
  return EAFNOSUPPORT;
}

/*****************************************************************************/
/*   ifpromisc:                                                              */
/*****************************************************************************/

/*
 *   We suppose that IFCONTROL_RCVPROMISCUOUS and IFCONTROL_RCVPROMISCOFF
 *   are undocumented commands for if_control applied to Ethernet driver
 *   in order to control the promiscuous mode of the driver.
 *   What I did is just to type "od -s /sdmach" and to look at the strings
 *   in the kernel and consider about them with manuals.
 */

int
ifpromisc(ifp, pswitch)
     struct ifnet *ifp;
     int pswitch;
{
  int ret;
  extern const char IFCONTROL_RCVPROMISCUOUS[];
  extern const char IFCONTROL_RCVPROMISCOFF[];

  bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc entered (pswitch = %d)\n",
	       pswitch, 0, 0, 0, 0);

  if (pswitch) {
    
    bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: calling IFCONTROL_RCVPROMISCUOUS ...\n",
		 0, 0, 0, 0, 0);
    
    ret = if_control((netif_t)ifp, IFCONTROL_RCVPROMISCUOUS, NULL);
    
    if (ret)
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: IFCONTROL_RCVPROMISCUOUS (error = %d).\n",
		   ret, 0, 0, 0, 0);
    else
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: IFCONTROL_RCVPROMISCOUS ok.\n",
		   0, 0, 0, 0, 0);
  }
  else {
    
    bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: calling IFCONTROL_RCVPROMISCOFF ...\n",
		 0, 0, 0, 0, 0);
    
    ret = if_control((netif_t)ifp, IFCONTROL_RCVPROMISCOFF, NULL);

    if (ret)
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: IFCONTROL_RCVPROMISCOFF (error = %d).\n",
		   ret, 0, 0, 0, 0);
    else
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifpromisc: IFCONTROL_RCVPROMISCOFF ok.\n",
		   0, 0, 0, 0, 0);
  }

  return ret;
}

/*****************************************************************************/
/*   ifmulticast:                                                            */
/*****************************************************************************/

/*
 *   We suppose that IFCONTROL_ADDMULTICAST and IFCONTROL_RMVMULTICAST
 *   are undocumented commands for if_control applied to the Ethernet driver
 *   in order to add and remove a multicast address.
 *   What I did is just to type "od -s /sdmach" and to look at the strings
 *   in the kernel and consider about them with manuals.
 */

int
ifmulticast(ifp, ifr, mswitch)
     struct ifnet *ifp;
     struct ifreq *ifr;
     int mswitch;
{
  int ret;
  extern const char IFCONTROL_ADDMULTICAST[];
  extern const char IFCONTROL_RMVMULTICAST[];
  u_char addr[6];

  bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast entered (addr = 0x%x, mswitch = %d)\n",
	       (int)addr, (int)mswitch, 0, 0, 0);

  /* Convert the given address in "ifr" into the Ethernet address in "addr". */
  switch (ifr->ifr_addr.sa_family) {
  case AF_UNSPEC:
    bcopy(ifr->ifr_addr.sa_data, addr, 6);
    {
      char s[64];
      sprintf(s, "%x:%x:%x:%x:%x:%x",
	      (int)addr[0], (int)addr[1], (int)addr[2],
	      (int)addr[3], (int)addr[4],  (int)addr[5]);
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: AF_UNSPEC (addr = %s)\n",
		   (int)s, 0, 0, 0, 0);
    }
    break;
  case AF_INET:
    {
      struct sockaddr_in *sin = (struct sockaddr_in *)&(ifr->ifr_addr);
      if (sin->sin_addr.s_addr == INADDR_ANY) {
	/*
	 * An IP address of INADDR_ANY means listen to all
	 * of the Ethernet multicast addresses for IP.
	 * (This is for the sake of IP muticast routers.)
	 */
	bpf_next_log(LOG_WARNING,
		     "ifmulticast: Sorry, AF_INET ofr INADDR_ANY is not supported yet.\n",
		     0, 0, 0, 0, 0);
	return EAFNOSUPPORT;
      }
      else {
	/* Map an IP multicast address to an Ethernet multicast address.
	   An Ethernet address has the length 6 bytes = 48 bits.
	   For an Ethernet multicast address, the high-order 25 bits
	   are fixed to 0x01005e and the low-order 23 bits are taken from
	   the low end of the corresponding IP address. */
	addr[0] = 0x01;
	addr[1] = 0x00;
	addr[2] = 0x5e;
	addr[3] = ((u_char *)&sin->sin_addr)[1] & 0x7f;
	addr[4] = ((u_char *)&sin->sin_addr)[2];
	addr[5] = ((u_char *)&sin->sin_addr)[3];
	/* Log the multicast address for debuging. */
	{
	  char s1[64], s2[64];
	  char *t = (char *)&sin->sin_addr;
	  sprintf(s1, "%d.%d.%d.%d",
		  (int)t[0], (int)t[1], (int)t[2], (int)t[3]);
	  sprintf(s2, "%x:%x:%x:%x:%x:%x",
		  (int)addr[0], (int)addr[1], (int)addr[2],
		  (int)addr[3], (int)addr[4],  (int)addr[5]);
	  bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: AF_INET (ip = %s, addr = %s)\n",
		       (int)s1, (int)s2, 0, 0, 0);
	}
      }
      break;
    }
  default:
    return EAFNOSUPPORT;
  }

  /* Which is the case, add or delete the multicast address ? */
  if (mswitch) {
    
    bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: calling IFCONTROL_ADDMULTICAST ...\n",
		 0, 0, 0, 0, 0);
    
    ret = if_control((netif_t)ifp, IFCONTROL_ADDMULTICAST, addr);
    
    if (ret)
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: IFCONTROL_ADDMULTICAST (error = %d).\n",
		   ret, 0, 0, 0, 0);
    else
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: IFCONTROL_ADDMULTICAST ok.\n",
		   0, 0, 0, 0, 0);
  }
  else {
    
    bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: calling IFCONTROL_RMVMULTICAST ...\n",
		 0, 0, 0, 0, 0);
    
    ret = if_control((netif_t)ifp, IFCONTROL_RMVMULTICAST, addr);

    if (ret)
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: IFCONTROL_RMVMULTICAST (error = %d).\n",
		   ret, 0, 0, 0, 0);
    else
      bpf_next_log(LOG_DEBUG_SPECIAL, "ifmulticast: IFCONTROL_RMVMULTICAST ok.\n",
		   0, 0, 0, 0, 0);
  }

  return ret;
}

/*****************************************************************************/
/*   bpf_next_log:                                                           */
/*****************************************************************************/

void
bpf_next_log(log_level, msg, arg1, arg2, arg3, arg4, arg5)
     int log_level;
     char *msg;
     int arg1;
     int arg2;
     int arg3;
     int arg4;
     int arg5;
{
#ifdef USE_PRINTF
  printf(msg, arg1, arg2, arg3, arg4, arg5);
#endif
#ifdef USE_KERN_SERV_LOG
  kern_serv_log(&instance, log_level, msg, arg1, arg2, arg3, arg4, arg5);
#endif
#ifdef USE_LOG
  log(log_level, msg, arg1, arg2, arg3, arg4, arg5);
#endif
}
